mod four;
mod game;
mod pl2;
use four::*;
use game::*;
use pl2::*;

use clap::Parser;
use std::ffi::{OsStr, OsString};
use std::io::{self, BufRead, BufReader, Write};
use std::mem::drop;
use std::process::{Child, ChildStdin, ChildStdout, Command, Stdio};

struct Bot {
    pub proc: Child,
    pub read: Option<BufReader<ChildStdout>>,
    pub write: Option<ChildStdin>,
}

impl Bot {
    fn new<S: AsRef<OsStr>>(program: S) -> io::Result<Self> {
        let mut proc = Command::new(program)
            .stdin(Stdio::piped())
            .stdout(Stdio::piped())
            .spawn()?;
        let read = proc.stdout.take().map(|s| BufReader::new(s));
        let write = proc.stdin.take();
        Ok(Self { proc, read, write })
    }
}

#[derive(Parser, Debug)]
#[command(author, version, about, long_about = None)]
struct Args {
    bot_x: OsString,
    bot_o: OsString,
}

fn main() -> io::Result<()> {
    let args = Args::parse();
    println!("Player X: {:#?}", &args.bot_x);
    println!("Player O: {:#?}", &args.bot_o);
    let mut bots = [Bot::new(&args.bot_x)?, Bot::new(&args.bot_o)?];
    if let Some(s) = &mut bots[0].write {
        s.write(b"f\n")?;
    }
    if let Some(s) = &mut bots[1].write {
        s.write(b"s\n")?;
    }
    let mut game: Four<BinaryPlayer> = Four::new();
    let mut line = String::new();
    let mut stdout = io::stdout();
    let mut bot_read = bots
        .iter_mut()
        .map(|bot| bot.read.take().expect("no stdin"))
        .collect::<Vec<_>>();
    let mut bot_write = bots
        .iter_mut()
        .map(|bot| bot.write.take().expect("no stdout"))
        .collect::<Vec<_>>();
    'iter: for (from, to) in [(0, 1), (1, 0)].into_iter().cycle() {
        let from = &mut bot_read[from];
        let to = &mut bot_write[to];
        let pl = game.current_player();
        match 'read: {
            line.clear();
            print!("waiting for {}... ", pl);
            stdout.flush()?;
            if from.read_line(&mut line).is_err() {
                println!("fail");
                break 'read None;
            }
            println!("output: {:?}", &line);
            to.write(&line.as_bytes())?;
            to.flush()?;
            match line.strip_suffix("\n").unwrap().parse::<u8>() {
                Ok(m) => FourMove::try_from(m).ok(),
                Err(_) => None,
            }
        } {
            Some(m) => {
                let out = match game.apply(m) {
                    Some(out) => {
                        print!("{}: {}.{}", pl, m, &game);
                        out
                    }
                    None => {
                        println!("{}: invalid move ({}).", pl, m);
                        println!("{} wins.", pl.next());
                        break 'iter;
                    }
                };
                match out {
                    MoveOutcome::Win(p) => {
                        println!("{} wins.", p);
                        break 'iter;
                    }
                    MoveOutcome::Tie => {
                        println!("Tie.");
                        break 'iter;
                    }
                    MoveOutcome::Pass => {
                        println!("...");
                    }
                }
            }
            None => {
                println!("{} generated invalid output: {:?}", pl, &line);
                println!("{} wins.", pl.next());
                break 'iter;
            }
        }
    }
    drop((bot_read, bot_write));
    let exits: Vec<_> = bots
        .into_iter()
        .map(|c| c.proc)
        .map(|mut c| c.wait())
        .collect();
    for e in exits {
        e?;
    }
    Ok(())
}
